attribute vec3 a_position;  // x, y : position( x, y )
                            //    z : height
                            
attribute vec2 a_normal;    // x, y : normal, 
attribute vec4 a_uv;        // x, y : diffuse texture uv for line body, 
                            // z, w : signed distance field uv
attribute vec4 a_line;      //    x : cap scale
                            //    y : distance ( meter )
                            //    z : 0 : stroke, 1: line body
                            //    w : width( meter )


uniform sampler2D u_diffuse_texture;

uniform mat4 u_mvp;
uniform vec4 u_scale;
uniform vec4 u_palette_info;    // x, y : per pixel uv of an atlas( pallete ) texture


                            
uniform vec4 u_view_info;   // x : distance per unit pixel 
                            // y : display scale

uniform vec4 u_view_level;  // x : view( map ) level ( int )
                            // y : view level ( flaot )
                            // z : pixel weight per int level

varying vec2 v_sdf_uv;           
varying vec2 v_dash_uv;         
varying vec4 v_diffuse;

varying vec4 v_line_info;   // x : boundary distance field value
                            // y : pixel distance
                            // z : original period
                            // w : period with width weight

const float sdf_weight = 0.01568247451;
const float int16_scale = 0.00006103515625;
void main()
{
    vec3 _position = a_position * u_scale.xyz;
    vec2 _normal = a_normal * int16_scale;
    vec4 _uv = a_uv * int16_scale;
    vec4 _line = vec4( a_line.xy, 0.0, 0.0 );
    
    
    // 스타일 팔레트에서 현재 카메라 z위치에 맞게 보간된, 라인 body 또는 stroke의 컬러값을 가져온다.
    float value_between_levels = ( u_view_level.y ) * u_palette_info.x;
    float value_body_or_storke = ( 1.0 - a_line.z ) * u_palette_info.y;
    vec2 uv = vec2( _uv.x + value_between_levels, _uv.y + value_body_or_storke );
    v_diffuse = texture2D( u_diffuse_texture, uv );
    
    // 스타일 팔레트에서 현재 카메라 z위치에 맞게 보간된, 라인 body 또는 stroke의 pixel width와 attribute에 있는 meter width중 큰 값을 가져온다.
    uv.y = _uv.y + 2.0 * u_palette_info.y;
    vec4 line_info = texture2D( u_diffuse_texture, uv );
    line_info.xy = line_info.xy * 255.0 *  u_view_info.y;
    float width = max( _line.w, line_info.x * u_view_info.x * 0.5 );
    float pixelWidth = width / u_view_info.x;
    float stroke_width = ( line_info.y * a_line.z ) * u_view_info.x;
    
    // line의 distance필드값을 만든다.
    vec2 pos_xy = _position.xy + _normal.xy  * ( ( width - stroke_width ) * _line.x );
    v_sdf_uv.xy = _uv.zw;
    v_line_info.x = max( 0.0, 1.0 - ( 1.0 / ( pixelWidth ) ) * u_view_info.y - sdf_weight ); 
    
    // pattern 정보를 계산한다.
    v_dash_uv.xy = vec2( _uv.x, _uv.y + ( 3.0 * u_palette_info.y ) );
    vec4 pattern_info = texture2D( u_diffuse_texture, v_dash_uv.xy );
    v_line_info.z = pattern_info.y * 255.0;
    v_line_info.y = ( pattern_info.x * 255.0 + _line.y ); 
//    v_pattern.z += v_pattern.z * u_view_level.y;;
    v_dash_uv.x = v_dash_uv.x + u_palette_info.x * 1.5;
    uv.x = _uv.x + ( u_view_level.x ) * u_palette_info.x;
    float line_width_at_int_level = texture2D( u_diffuse_texture, uv ).x;
    line_width_at_int_level = line_width_at_int_level * 255.0 *  u_view_info.y;
    v_line_info.w = max( 0.01, line_width_at_int_level ) * u_view_level.z;
    
    
    gl_Position = u_mvp * vec4( pos_xy / u_scale.xy, a_position.z , 1.0 );
}
